#include <stdio.h>
#include <string.h>
#include <stdlib.h>
//#include "ima.h"
#include "imalib.h"
#include "hardware.h"

extern const __y s_int16 imaDownSampleBy3[102];
extern const __y s_int16 imaDownSampleBy2[64];

__align s_int16 upBuf[34];
//s_int16 *upPtr = upBuf;
//s_int16 upPhase = 0;

/*
  0 1 2 3 4 5 6 7 8
  x 0 0 0 0 0 0 0 0 phase = 0
  0 x 0 0 0 0 0 0 0 phase = 1
  0 0 x 0 0 0 0 0 0 phase = 2
  y 0 0 x 0 0 0 0 0 phase = 0
  0 y 0 0 x 0 0 0 0 phase = 1
  0 0 y 0 0 x 0 0 0 phase = 2
  z 0 0 y 0 0 x 0 0 phase = 0
  0 z 0 0 y 0 0 x 0 phase = 1
  0 0 z 0 0 y 0 0 x phase = 2
  ..
 */
#define INTERP_NEED_NEW_SAMPLE(ip) ((ip)->phase==0)
#define INTERP_SET_NEW_SAMPLE(ip,s) (*((ip)->ptr) = (s))
struct INTERPOLATION {
    s_int16 phase; /* if phase==0, write new sample to *ptr */
    s_int16 *ptr;
    s_int16 *buf;  /* must be aligned according to bufSize */
    s_int16 bufSize;
    s_int16 ratio; /* interpolation ratio. bufSize >= n/ratio */
    s_int16 n;     /* number of filter coefficients, n%ratio == 0 */
    __y s_int16 *coeff; /* filter coefficients */
    s_int16 loops; /* n/ratio */
    //s_int16 getSampleParam;
    //s_int16 (*GetSample)(s_int16 param); /* called when new sample needed */
};
#define INTERP_PHASE   0
#define INTERP_PTR     1
#define INTERP_BUF     2
#define INTERP_BUFSIZE 3
#define INTERP_RATIO   4
#define INTERP_N       5
#define INTERP_COEFF   6
#define INTERP_LOOPS   7

void InterpolationInit(register __i2 struct INTERPOLATION *ip,
			  register __a0 s_int16 ratio,
			  register __i0 s_int16 *buf,
			  register __a1 s_int16 bufSize,
			  register __i1 __y s_int16 *coeff,
			  register __c0 s_int16 n) {
    ip->phase = 0;
    ip->ptr = buf;
    ip->buf = buf;
    ip->bufSize = bufSize;
    ip->ratio = ratio;
    ip->n = n;
    ip->coeff = coeff;
    ip->loops = n/ratio;
}

#ifdef __VSDSP__
/* asm version 67.3 cycles for interpolate by 3, filter length 102.
   -> 1.6MHz at 24000Hz (8k->24k) */
s_int16 InterpolationFilter(register __i2 struct INTERPOLATION *ip);
#else
s_int16 InterpolationFilter(register __i2 struct INTERPOLATION *ip) {
    s_int16 out;
    {
	register int i;
	register s_int16 *p = ip->ptr;
	register __y s_int16 *c = ip->coeff+ip->phase;
	register s_int32 accu = 0;
	s_int16 l = ip->loops;
	for (i=0; i<l; i++) {
	    accu += (s_int32)*p++ * *c;
	    if (p >= ip->buf+ip->bufSize)
		p -= ip->bufSize;
	    c += ip->ratio;
	}
	accu >>= 15;
	if (accu > 32767) {
	    accu = 32767;
	} else if (accu < -32768) {
	    accu = -32768;
	}
	out = accu;
    }
    if (++ip->phase >= ip->ratio) {
	ip->phase = 0;
	if (--ip->ptr < ip->buf)
	    ip->ptr += ip->bufSize;
    }
    return out;
}
#endif

#if 0
s_int16 PcmFilter(s_int16 in, s_int16 *out) { /*900 cycles*/
    int i;
    register s_int16 *p = upPtr;
    s_int32 accu = 0;
    for (i=upPhase;i<102;i+=3) {
	accu += (s_int32)*p++ * imaDownSampleBy3[i];
	if (p >= upBuf+sizeof(upBuf))
	    p -= sizeof(upBuf);
    }
    accu >>= 15;
    if (accu > 32767) {
	accu = 32767;
    } else if (accu < -32768) {
	accu = -32768;
    }
    *out = accu >> 15;
    if (++upPhase==3) {
	upPhase = 0;
	--upPtr;
	if (upPtr < upBuf)
	    upPtr += sizeof(upBuf);
	*upPtr = in;
	return 1; /*need new sample*/
    }
    return 0;
}
#endif

struct INTERPOLATION ip3;
u_int16 pcmFreeGPIO;

#ifdef SIM
void MyMain(void) {
    FILE *fpo = fopen("output.dat", "wb");
    FILE *fp = fopen("input.dat", "rb");
    printf("test\n");
    USEX(INT_ENABLE) = 0;
#if 0
    InterpolationInit(&ip3, 3, upBuf, sizeof(upBuf), imaDownSampleBy3, 102);
    if (fp) {
	char line[80];
	while (!feof(fp)) {
	    s_int16 out;
	    if (INTERP_NEED_NEW_SAMPLE(&ip3)) {
		if (fgets(line, 80, fp)) {
		    s_int16 sample = strtol(line, NULL, 0);
//printf("adding %d\n", sample);
		    INTERP_SET_NEW_SAMPLE(&ip3, sample);
		} else {
		    break;
		}
	    }
	    out = InterpolationFilter(&ip3);
	    if (fpo)
		fprintf(fpo, "%d\n", out);
	}
    }

#else
    memset(upBuf, 0, sizeof(upBuf));
    upPtr = upBuf;
    if (fp) {
	char line[80];
	s_int16 sample = 0;
	while (!feof(fp)) {
	    s_int16 out;
	    s_int16 n = PcmFilter(sample, &out);
	    if (fpo)
		fprintf(fpo, "%d\n", out);
	    if (n) {
		if (fgets(line, 80, fp)) {
		    sample = strtol(line, NULL, 0);
		} else {
		    break;
		}
	    }
	}
    }
#endif
    fclose(fp);
    fclose(fpo);
}
#else /*SIM*/

#include <general.h>
#define PCM_FIFO_SIZE 256 //2048 //1024
__align __y s_int16 pcmFifo[PCM_FIFO_SIZE];
volatile __y s_int16 * __x pcmFifoRd, * __x pcmFifoWr;
s_int16 pcmVol;
s_int16 offDelay;

auto u_int16 DB12(register __b0 s_int16 n);
void SetPcmFree(void) {
    register s_int16 free = pcmFifoRd - pcmFifoWr;
    if (free <= 0)
	free += PCM_FIFO_SIZE;
    if (free < 3*PCM_FIFO_SIZE/4) {
	offDelay = 0;
	USEX(INT_ENABLE) |= (1<<INT_EN_SDM);
    }
#if 1
    if (pcmFreeGPIO) {
	if (free > 32) {
	    USEX(GPIO_ODATA) |= pcmFreeGPIO;
	} else {
	    USEX(GPIO_ODATA) &= ~pcmFreeGPIO;
	}
    }
#endif
    USEX(SCI_AICTRL0) = free-1;
}
s_int16 ApplAddr(s_int16 register __i0 **d,
		 s_int16 register __a1 mode,
		 s_int16 register __a0 n) {
    if (mode == APPL_W1) {
	*pcmFifoWr = USEX(SCI_AICTRL1);
	if (++pcmFifoWr >= &pcmFifo[PCM_FIFO_SIZE])
	    pcmFifoWr -= PCM_FIFO_SIZE;
	SetPcmFree();
    } else if (mode == APPL_R0) {
	SetPcmFree();
    } else if (mode == APPL_W2) {
	register u_int16 vol = USEX(SCI_VOL);
	register u_int16 pvol = USEX(SCI_AICTRL2) & 255;
	if ((vol & 255) < 4 || (vol & 0xff00) < 0x0400) {
	    vol = 0x0404;
	    USEX(SCI_VOL) = vol;
	}
	if ((vol & 255) + (pvol & 255) < 12) {
	    pvol = 12 - (vol & 255);
	    USEX(SCI_AICTRL2) = pvol;
	}
	/* Already -6dB in volume multiplication, -10.5dB total */
	pcmVol = DB12(9+pvol); /*works from 0..191*/
    }
    return n;
}
#include "asmfuncs.h"
void SDMInt(void) {
    register s_int16 data = *pcmFifoRd;
    Enable();
#if 1
    if (INTERP_NEED_NEW_SAMPLE(&ip3)) {
	register __y s_int16 *p = pcmFifoRd;
	data = *p;
	INTERP_SET_NEW_SAMPLE(&ip3, data);
	if (++p >= &pcmFifo[PCM_FIFO_SIZE])
	    p -= PCM_FIFO_SIZE;
	if (p == pcmFifoWr) {
	    /* empty -- if persists for 0.125s, turn processing off. */
//USEX(SCI_AICTRL3)++;
	    if (++offDelay > 1000) {
		USEX(INT_ENABLE) &= ~(1<<INT_EN_SDM);
		USEX(SDM_DATA_RIGHT) = USEX(SDM_DATA_LEFT) = 0;
		Disable();
		return;
	    }
	    *pcmFifoRd = (f_int16)data * (f_int16)0.99;
	} else {
	    offDelay = 0;
	    pcmFifoRd = p;
	    //PcmSetFree();
	    SetPcmFree();
	}
    }
    if (ip3.ratio > 1) {
	data = InterpolationFilter(&ip3);
    }
#ifndef PCM_LEFT
    USEX(SDM_DATA_RIGHT) =
#endif
#ifndef PCM_RIGHT
      USEX(SDM_DATA_LEFT) =
#endif
        pcmVol * (s_int32)data >> 16;
    Disable();
#else
    if (INTERP_NEED_NEW_SAMPLE(&ip3)) {
	INTERP_SET_NEW_SAMPLE(&ip3, data);
	if (pcmFifoRd == pcmFifoWr) {
	    /* empty -- if persists for 0.125s, turn processing off. */
//USEX(SCI_AICTRL3)++;
	    if (++offDelay > 1000) { /*20110322 fixed < to >!! */
		USEX(INT_ENABLE) &= ~(1<<INT_EN_SDM);
		USEX(SDM_DATA_RIGHT) = USEX(SDM_DATA_LEFT) = 0;
	    }
	    *pcmFifoRd = (f_int16)data * (f_int16)0.999;
	} else {
	    offDelay = 0;
	    if (++pcmFifoRd >= &pcmFifo[PCM_FIFO_SIZE])
		pcmFifoRd -= PCM_FIFO_SIZE;
	    SetPcmFree();
	}
    }
    if (ip3.ratio > 1) {
	data = InterpolationFilter(&ip3);
    }
    USEX(SDM_DATA_RIGHT) = USEX(SDM_DATA_LEFT) = pcmVol * (s_int32)data >> 16;
    Disable();
#endif
}

void MyMain(void) {
    pcmFreeGPIO = USEX(SCI_AICTRL1); /* which GPIO to use for free state? */
    pcmFifoRd = pcmFifoWr = pcmFifo;
    memset(upBuf, 0, sizeof(upBuf));
#if 0
    /* Configure SDM output to correct rate. */
    USEX(SDM_CONTROL) = SDM_ENABLE | (256-1); //24kHz
    if (USEX(SCI_AICTRL0) == 24000) { /* as-is */
	InterpolationInit(&ip3, 1, upBuf, sizeof(upBuf), NULL, 0);
    } else if (USEX(SCI_AICTRL0) == 16000) {
	USEX(SDM_CONTROL) = SDM_ENABLE | (192-1); //32kHz
	InterpolationInit(&ip3, 2, upBuf, sizeof(upBuf), imaDownSampleBy2, 64);
    } else if (USEX(SCI_AICTRL0) == 12000) {
	/*Middle quality*/
	InterpolationInit(&ip3, 2, upBuf, sizeof(upBuf), imaDownSampleBy2, 64);
    } else {
	/*Default: 8kHz*/
	InterpolationInit(&ip3, 3, upBuf, sizeof(upBuf), imaDownSampleBy3,102);
    }
#else
    /* From VS1063A */
    if (USEX(SCI_AICTRL0) >= 22050U) {
	/*High quality -> as-is */
	InterpolationInit(&ip3, 1, upBuf, sizeof(upBuf), NULL, 0);
    } else if (USEX(SCI_AICTRL0) >= 11025) {
	/*Middle quality -> up by 2*/
	InterpolationInit(&ip3, 2, upBuf, sizeof(upBuf),
			  imaDownSampleBy2, sizeof(imaDownSampleBy2));
    } else {
	/*Default: 8kHz -> up by 3*/
	InterpolationInit(&ip3, 3, upBuf, sizeof(upBuf),
			  imaDownSampleBy3, sizeof(imaDownSampleBy3));
    }
    {
	register s_int16 sdmDiv =
	    (s_int32)extClock4KHz * 2000 /
	    ((s_int32)USEX(SCI_AICTRL0) * ip3.ratio);
	USEX(SDM_CONTROL) = SDM_ENABLE | (sdmDiv-1);
    }
#endif

    USEX(GPIO_DDR) |= pcmFreeGPIO;
    SetPcmFree();
    applAddr = ApplAddr;
    ApplAddr(NULL, APPL_W2, 0); //Update pcmVol
}
#endif /*elseSIM*/
